@@ -1,27 +1,13 @@
-/*  SPDX-License-Identifier: LGPL-2.0-or-later
+/*
+    SPDX-FileCopyrightText: 2002, 2003 Anders Lund <anders.lund@lund.tdcadsl.dk>
+    SPDX-FileCopyrightText: 2003 Christoph Cullmann <cullmann@kde.org>
+    SPDX-FileCopyrightText: 2001 Joseph Wenninger <jowenn@kde.org>
+    SPDX-FileCopyrightText: 2006-2016 Dominik Haumann <dhaumann@kde.org>
+    SPDX-FileCopyrightText: 2007 Mirko Stocker <me@misto.ch>
+    SPDX-FileCopyrightText: 2009 Michel Ludwig <michel.ludwig@kdemail.net>
+    SPDX-FileCopyrightText: 1999 Jochen Wilhelmy <digisnap@cs.tu-berlin.de>
 
-    Copyright (C) 2002, 2003 Anders Lund <anders.lund@lund.tdcadsl.dk>
-    Copyright (C) 2003 Christoph Cullmann <cullmann@kde.org>
-    Copyright (C) 2001 Joseph Wenninger <jowenn@kde.org>
-    Copyright (C) 2006-2016 Dominik Haumann <dhaumann@kde.org>
-    Copyright (C) 2007 Mirko Stocker <me@misto.ch>
-    Copyright (C) 2009 Michel Ludwig <michel.ludwig@kdemail.net>
-    Copyright (C) 1999 Jochen Wilhelmy <digisnap@cs.tu-berlin.de>
-
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Library General Public
-    License as published by the Free Software Foundation; either
-    version 2 of the License, or (at your option) any later version.
-
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Library General Public License for more details.
-
-    You should have received a copy of the GNU Library General Public License
-    along with this library; see the file COPYING.LIB.  If not, write to
-    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-    Boston, MA 02110-1301, USA.
+    SPDX-License-Identifier: LGPL-2.0-or-later
 */
 
 #ifndef KATE_DIALOGS_H
@@ -63,11 +49,8 @@ class Job;
 class TransferJob;
 }
 
-class KComboBox;
 class KShortcutsEditor;
 class QSpinBox;
-class KPluginSelector;
-class KPluginInfo;
 class KProcess;
 
 class QCheckBox;
@@ -125,7 +108,7 @@ class KateDictionaryBar : public KateViewBarWidget
 
 public:
     explicit KateDictionaryBar(KTextEditor::ViewPrivate *view, QWidget *parent = nullptr);
-    virtual ~KateDictionaryBar();
+    ~KateDictionaryBar() override;
 
 public Q_SLOTS:
     void updateData();
@@ -234,6 +217,8 @@ public:
 private:
     Ui::NavigationConfigWidget *ui;
 
+    void initMulticursorModifierComboBox();
+
 public Q_SLOTS:
     void apply() override;
     void reload() override;
@@ -356,7 +341,7 @@ private:
  * This dialog will prompt the user for what do with a file that is
  * modified on disk.
  * If the file wasn't deleted, it has a 'diff' button, which will create
- * a diff file (using diff(1)) and launch that using KRun.
+ * a diff file (using diff(1)) and launch that using OpenUrlJob.
  */
 class KateModOnHdPrompt : public QObject
 {
@@ -370,7 +355,7 @@ public:
         Close
     };
     KateModOnHdPrompt(KTextEditor::DocumentPrivate *doc, KTextEditor::ModificationInterface::ModifiedOnDiskReason modtype, const QString &reason);
-    ~KateModOnHdPrompt();
+    ~KateModOnHdPrompt() override;
 
 Q_SIGNALS:
     void saveAsTriggered();
@@ -387,12 +372,13 @@ private Q_SLOTS:
 
 private Q_SLOTS:
     void slotDataAvailable(); ///< read data from the process
-    void slotPDone();         ///< Runs the diff file when done
+    void slotPDone(); ///< Runs the diff file when done
 
 private:
     KTextEditor::DocumentPrivate *m_doc;
     QPointer<KTextEditor::Message> m_message;
     KTextEditor::ModificationInterface::ModifiedOnDiskReason m_modtype;
+    QString m_fullDiffPath;
     KProcess *m_proc;
     QTemporaryFile *m_diffFile;
     QAction *m_diffAction;
