@@ -27,6 +27,7 @@
 
 #include <QBrush>
 #include <QPainter>
+#include <QPainterPath>
 #include <QRegularExpression>
 #include <QStack>
 #include <QtMath> // qCeil
@@ -139,20 +140,20 @@ void KateRenderer::setShowSelections(bool showSelections)
     m_showSelections = showSelections;
 }
 
-void KateRenderer::increaseFontSizes(qreal step)
+void KateRenderer::increaseFontSizes(qreal step) const
 {
     QFont f(config()->baseFont());
     f.setPointSizeF(f.pointSizeF() + step);
     config()->setFont(f);
 }
 
-void KateRenderer::resetFontSizes()
+void KateRenderer::resetFontSizes() const
 {
     QFont f(KateRendererConfig::global()->baseFont());
     config()->setFont(f);
 }
 
-void KateRenderer::decreaseFontSizes(qreal step)
+void KateRenderer::decreaseFontSizes(qreal step) const
 {
     QFont f(config()->baseFont());
     if ((f.pointSizeF() - step) > 0) {
@@ -241,7 +242,7 @@ void KateRenderer::paintTextLineBackground(QPainter &paint, KateLineLayoutPtr la
     }
 }
 
-void KateRenderer::paintTabstop(QPainter &paint, qreal x, qreal y)
+void KateRenderer::paintTabstop(QPainter &paint, qreal x, qreal y) const
 {
     QPen penBackup(paint.pen());
     QPen pen(config()->tabMarkerColor());
@@ -263,7 +264,7 @@ void KateRenderer::paintTabstop(QPainter &paint, qreal x, qreal y)
     paint.setPen(penBackup);
 }
 
-void KateRenderer::paintSpace(QPainter &paint, qreal x, qreal y)
+void KateRenderer::paintSpace(QPainter &paint, qreal x, qreal y) const
 {
     QPen penBackup(paint.pen());
     QPen pen(config()->tabMarkerColor());
@@ -277,7 +278,7 @@ void KateRenderer::paintSpace(QPainter &paint, qreal x, qreal y)
     paint.setRenderHint(QPainter::Antialiasing, false);
 }
 
-void KateRenderer::paintNonBreakSpace(QPainter &paint, qreal x, qreal y)
+void KateRenderer::paintNonBreakSpace(QPainter &paint, qreal x, qreal y) const
 {
     QPen penBackup(paint.pen());
     QPen pen(config()->tabMarkerColor());
@@ -520,7 +521,9 @@ QVector<QTextLayout::FormatRange> KateRenderer::decorationsForLine(const Kate::T
             backgroundAttribute = KTextEditor::Attribute::Ptr(new KTextEditor::Attribute());
         }
 
-        backgroundAttribute->setBackground(config()->selectionColor());
+        if (!hasCustomLineHeight()) {
+            backgroundAttribute->setBackground(config()->selectionColor());
+        }
         backgroundAttribute->setForeground(attribute(KTextEditor::dsNormal)->selectedForeground().color());
 
         // Create a range for the current selection
@@ -555,6 +558,10 @@ QVector<QTextLayout::FormatRange> KateRenderer::decorationsForLine(const Kate::T
         endPosition = KTextEditor::Cursor(line + 1, 0);
     }
 
+    // Background formats have lower priority so they get overriden by selection
+    const bool shoudlClearBackground = hasCustomLineHeight() && m_view->selection();
+    const KTextEditor::Range selectionRange = m_view->selectionRange();
+
     // Main iterative loop.  This walks through each set of highlighting ranges, and stops each
     // time the highlighting changes.  It then creates the corresponding QTextLayout::FormatRanges.
     QVector<QTextLayout::FormatRange> newHighlight;
@@ -591,6 +598,11 @@ QVector<QTextLayout::FormatRange> KateRenderer::decorationsForLine(const Kate::T
             }
         }
 
+        // Clear background if this position overlaps selection
+        if (shoudlClearBackground && fr.format.hasProperty(QTextFormat::BackgroundBrush) && selectionRange.overlapsColumn(currentPosition.column())) {
+            fr.format.clearBackground();
+        }
+
         newHighlight.append(fr);
 
         currentPosition = nextPosition;
@@ -606,7 +618,7 @@ QVector<QTextLayout::FormatRange> KateRenderer::decorationsForLine(const Kate::T
     return newHighlight;
 }
 
-void KateRenderer::assignSelectionBrushesFromAttribute(QTextLayout::FormatRange &target, const KTextEditor::Attribute &attribute) const
+void KateRenderer::assignSelectionBrushesFromAttribute(QTextLayout::FormatRange &target, const KTextEditor::Attribute &attribute)
 {
     if (attribute.hasProperty(SelectedForeground)) {
         target.format.setForeground(attribute.selectedForeground());
@@ -616,6 +628,42 @@ void KateRenderer::assignSelectionBrushesFromAttribute(QTextLayout::FormatRange
     }
 }
 
+void KateRenderer::paintTextLineSelection(QPainter &paint, KateLineLayoutPtr layout, const QVector<QTextLayout::FormatRange> &selRanges)
+{
+    QPainterPath p;
+    for (const auto &sel : selRanges) {
+        const int s = sel.start;
+        const int e = sel.start + sel.length;
+
+        const int startViewLine = layout->viewLineForColumn(s);
+        const int endViewLine = layout->viewLineForColumn(e);
+        if (startViewLine == endViewLine) {
+            KateTextLayout l = layout->viewLine(startViewLine);
+            const int startX = cursorToX(l, s);
+            const int endX = cursorToX(l, e);
+            const int y = startViewLine * lineHeight();
+            QRect r(startX, y, (endX - startX), lineHeight());
+            p.addRect(r);
+        } else {
+            for (int l = startViewLine; l <= endViewLine; ++l) {
+                auto kateLayout = layout->viewLine(l);
+                int sx = 0;
+                int width = kateLayout.lineLayout().naturalTextWidth();
+                if (l == startViewLine) {
+                    sx = kateLayout.lineLayout().cursorToX(s);
+                } else if (l == endViewLine) {
+                    width = kateLayout.lineLayout().cursorToX(e);
+                }
+
+                const int y = l * lineHeight();
+                QRect r(sx, y, width - sx, lineHeight());
+                p.addRect(r);
+            }
+        }
+    }
+    paint.fillPath(p, config()->selectionColor());
+}
+
 void KateRenderer::paintTextLine(QPainter &paint, KateLineLayoutPtr range, int xStart, int xEnd, const KTextEditor::Cursor *cursor, PaintTextLineFlags flags)
 {
     Q_ASSERT(range->isValid());
@@ -678,6 +726,9 @@ void KateRenderer::paintTextLine(QPainter &paint, KateLineLayoutPtr range, int x
             // Draw the text :)
             if (drawSelection) {
                 additionalFormats = decorationsForLine(range->textLine(), range->line(), true);
+                if (hasCustomLineHeight()) {
+                    paintTextLineSelection(paint, range, additionalFormats);
+                }
                 range->layout()->draw(&paint, QPoint(-xStart, 0), additionalFormats);
 
             } else {
@@ -794,83 +845,23 @@ void KateRenderer::paintTextLine(QPainter &paint, KateLineLayoutPtr range, int x
             }
         }
 
-        // Draw caret
-        if (drawCaret() && cursor && range->includesCursor(*cursor)) {
-            int caretWidth;
-            int lineWidth = 2;
-            QColor color;
-            QTextLine line = range->layout()->lineForTextPosition(qMin(cursor->column(), range->length()));
-
-            // Determine the caret's style
-            caretStyles style = caretStyle();
-
-            // Make the caret the desired width
-            if (style == Line) {
-                caretWidth = lineWidth;
-            } else if (line.isValid() && cursor->column() < range->length()) {
-                caretWidth = int(line.cursorToX(cursor->column() + 1) - line.cursorToX(cursor->column()));
-                if (caretWidth < 0) {
-                    caretWidth = -caretWidth;
-                }
-            } else {
-                caretWidth = spaceWidth();
-            }
-
-            // Determine the color
-            if (m_caretOverrideColor.isValid()) {
-                // Could actually use the real highlighting system for this...
-                // would be slower, but more accurate for corner cases
-                color = m_caretOverrideColor;
-            } else {
-                // search for the FormatRange that includes the cursor
-                const auto formatRanges = range->layout()->formats();
-                for (const QTextLayout::FormatRange &r : formatRanges) {
-                    if ((r.start <= cursor->column()) && ((r.start + r.length) > cursor->column())) {
-                        // check for Qt::NoBrush, as the returned color is black() and no invalid QColor
-                        QBrush foregroundBrush = r.format.foreground();
-                        if (foregroundBrush != Qt::NoBrush) {
-                            color = r.format.foreground().color();
-                        }
-                        break;
-                    }
-                }
-                // still no color found, fall back to default style
-                if (!color.isValid()) {
-                    color = attribute(KTextEditor::dsNormal)->foreground().color();
+        // Draw carets
+        const auto &secCursors = view()->secondaryCursors();
+        // Find carets on this line
+        auto mIt = std::lower_bound(secCursors.begin(), secCursors.end(), range->line(), [](const KTextEditor::ViewPrivate::SecondaryCursor &l, int line) {
+            return l.pos->line() < line;
+        });
+        if (mIt != secCursors.end() && mIt->cursor().line() == range->line()) {
+            for (; mIt != secCursors.end(); ++mIt) {
+                auto cursor = mIt->cursor();
+                if (cursor.line() == range->line()) {
+                    paintCaret(&cursor, range, paint, xStart, xEnd);
+                } else {
+                    break;
                 }
             }
-
-            paint.save();
-            switch (style) {
-            case Line:
-                paint.setPen(QPen(color, caretWidth));
-                break;
-            case Block:
-                // use a gray caret so it's possible to see the character
-                color.setAlpha(128);
-                paint.setPen(QPen(color, caretWidth));
-                break;
-            case Underline:
-                break;
-            case Half:
-                color.setAlpha(128);
-                paint.setPen(QPen(color, caretWidth));
-                break;
-            }
-
-            if (cursor->column() <= range->length()) {
-                range->layout()->drawCursor(&paint, QPoint(-xStart, 0), cursor->column(), caretWidth);
-            } else {
-                // Off the end of the line... must be block mode. Draw the caret ourselves.
-                const KateTextLayout &lastLine = range->viewLine(range->viewLineCount() - 1);
-                int x = cursorToX(lastLine, KTextEditor::Cursor(range->line(), cursor->column()), true);
-                if ((x >= xStart) && (x <= xEnd)) {
-                    paint.fillRect(x - xStart, (int)lastLine.lineLayout().y(), caretWidth, lineHeight(), color);
-                }
-            }
-
-            paint.restore();
         }
+        paintCaret(cursor, range, paint, xStart, xEnd);
     }
 
     // show word wrap marker if desirable
@@ -916,6 +907,90 @@ void KateRenderer::paintTextLine(QPainter &paint, KateLineLayoutPtr range, int x
     }
 }
 
+void KateRenderer::paintCaret(const KTextEditor::Cursor *cursor, const KateLineLayoutPtr &range, QPainter &paint, int xStart, int xEnd)
+{
+    if (drawCaret() && cursor && range->includesCursor(*cursor)) {
+        int caretWidth;
+        int lineWidth = 2;
+        QColor color;
+        QTextLine line = range->layout()->lineForTextPosition(qMin(cursor->column(), range->length()));
+
+        // Determine the caret's style
+        caretStyles style = caretStyle();
+
+        // Make the caret the desired width
+        if (style == Line) {
+            caretWidth = lineWidth;
+        } else if (line.isValid() && cursor->column() < range->length()) {
+            caretWidth = int(line.cursorToX(cursor->column() + 1) - line.cursorToX(cursor->column()));
+            if (caretWidth < 0) {
+                caretWidth = -caretWidth;
+            }
+        } else {
+            caretWidth = spaceWidth();
+        }
+
+        // Determine the color
+        if (m_caretOverrideColor.isValid()) {
+            // Could actually use the real highlighting system for this...
+            // would be slower, but more accurate for corner cases
+            color = m_caretOverrideColor;
+        } else {
+            // search for the FormatRange that includes the cursor
+            const auto formatRanges = range->layout()->formats();
+            for (const QTextLayout::FormatRange &r : formatRanges) {
+                if ((r.start <= cursor->column()) && ((r.start + r.length) > cursor->column())) {
+                    // check for Qt::NoBrush, as the returned color is black() and no invalid QColor
+                    QBrush foregroundBrush = r.format.foreground();
+                    if (foregroundBrush != Qt::NoBrush) {
+                        color = r.format.foreground().color();
+                    }
+                    break;
+                }
+            }
+            // still no color found, fall back to default style
+            if (!color.isValid()) {
+                color = attribute(KTextEditor::dsNormal)->foreground().color();
+            }
+        }
+
+        paint.save();
+        switch (style) {
+        case Line:
+            paint.setPen(QPen(color, caretWidth));
+            break;
+        case Block:
+            // use a gray caret so it's possible to see the character
+            color.setAlpha(128);
+            paint.setPen(QPen(color, caretWidth));
+            break;
+        case Underline:
+            break;
+        case Half:
+            color.setAlpha(128);
+            paint.setPen(QPen(color, caretWidth));
+            break;
+        }
+
+        if (cursor->column() <= range->length()) {
+            // Ensure correct cursor placement for RTL text
+            if (range->layout()->textOption().textDirection() == Qt::RightToLeft) {
+                xStart += caretWidth;
+            }
+            range->layout()->drawCursor(&paint, QPoint(-xStart, 0), cursor->column(), caretWidth);
+        } else {
+            // Off the end of the line... must be block mode. Draw the caret ourselves.
+            const KateTextLayout &lastLine = range->viewLine(range->viewLineCount() - 1);
+            int x = cursorToX(lastLine, KTextEditor::Cursor(range->line(), cursor->column()), true);
+            if ((x >= xStart) && (x <= xEnd)) {
+                paint.fillRect(x - xStart, (int)lastLine.lineLayout().y(), caretWidth, lineHeight(), color);
+            }
+        }
+
+        paint.restore();
+    }
+}
+
 uint KateRenderer::fontHeight() const
 {
     return m_fontHeight;
@@ -982,6 +1057,11 @@ void KateRenderer::updateConfig()
     }
 }
 
+bool KateRenderer::hasCustomLineHeight() const
+{
+    return !qFuzzyCompare(config()->lineHeightMultiplier(), 1.0);
+}
+
 void KateRenderer::updateFontHeight()
 {
     // cache font + metrics
@@ -1002,6 +1082,15 @@ void KateRenderer::updateFontHeight()
     // qreal fontHeight = font.ascent() + font.descent();
     m_fontHeight = qMax(1, qCeil(m_fontMetrics.ascent() + m_fontMetrics.descent()));
     m_fontAscent = m_fontMetrics.ascent();
+
+    if (hasCustomLineHeight()) {
+        const auto oldFontHeight = m_fontHeight;
+        const qreal newFontHeight = qreal(m_fontHeight) * config()->lineHeightMultiplier();
+        m_fontHeight = newFontHeight;
+
+        qreal diff = std::abs(oldFontHeight - newFontHeight);
+        m_fontAscent += (diff / 2);
+    }
 }
 
 void KateRenderer::updateMarkerSize()
@@ -1056,6 +1145,13 @@ void KateRenderer::layoutLine(KateLineLayoutPtr lineLayout, int maxwidth, bool c
     if (isLineRightToLeft(lineLayout)) {
         opt.setAlignment(Qt::AlignRight);
         opt.setTextDirection(Qt::RightToLeft);
+        // Must turn off this flag otherwise cursor placement
+        // is totally broken.
+        if (view()->config()->dynWordWrap()) {
+            auto flags = opt.flags();
+            flags.setFlag(QTextOption::IncludeTrailingSpaces, false);
+            opt.setFlags(flags);
+        }
     } else {
         opt.setAlignment(Qt::AlignLeft);
         opt.setTextDirection(Qt::LeftToRight);
@@ -1149,7 +1245,7 @@ void KateRenderer::layoutLine(KateLineLayoutPtr lineLayout, int maxwidth, bool c
 // 3) QString::isRightToLeft() does not seem to work on my setup
 // 4) isStringRightToLeft() should behave much better than QString::isRightToLeft() therefore:
 // 5) isStringRightToLeft() kicks ass
-bool KateRenderer::isLineRightToLeft(KateLineLayoutPtr lineLayout) const
+bool KateRenderer::isLineRightToLeft(KateLineLayoutPtr lineLayout)
 {
     QString s = lineLayout->textLine()->text();
     int i = 0;
